@@ -11,8 +11,6 @@ mkdir abi-compliance/dumps
 # build the current changes
 export SKIP_TESTS=ON
 export EXTRA_CONFIGURE_FLAGS="-DCMAKE_INSTALL_PREFIX=./abi-compliance/changes-install -DCMAKE_C_FLAGS=-g -Og"
-# Use GitPython from venv.
-. venv/bin/activate
 python build/calc_release_version.py > VERSION_CURRENT
 python build/calc_release_version.py -p > VERSION_RELEASED
 sh .evergreen/compile.sh
