@@ -17,9 +17,7 @@
 
 #include <bson/bson.h>
 #include <bson/bcon.h>
-#define BSON_INSIDE
 #include <bson/bson-private.h>
-#undef BSON_INSIDE
 #include <fcntl.h>
 #include <time.h>
 
