@@ -46,6 +46,10 @@ _is_resumable_error (const bson_t *reply)
       return true;
    }
 
+   if (mongoc_error_has_label (reply, "NonRetryableChangeStreamError")) {
+      return false;
+   }
+
    /* Change Streams Spec resumable criteria: "a server error response with an
     * error message containing the substring 'not master' or 'node is
     * recovering' */
@@ -66,6 +70,19 @@ _is_resumable_error (const bson_t *reply)
       return true;
    }
 }
+
+
+static void
+_set_resume_token (mongoc_change_stream_t *stream, const bson_t *resume_token)
+{
+   BSON_ASSERT (stream);
+   BSON_ASSERT (resume_token);
+
+   bson_destroy (&stream->resume_token);
+   bson_copy_to (resume_token, &stream->resume_token);
+}
+
+
 /* construct the aggregate command in cmd. looks like one of the following:
  * for a collection change stream:
  *   { aggregate: collname, pipeline: [], cursor: { batchSize: x } }
@@ -76,7 +93,9 @@ _is_resumable_error (const bson_t *reply)
  *     cursor: { batchSize: x } }
  */
 static void
-_make_command (mongoc_change_stream_t *stream, bson_t *command)
+_make_command (mongoc_change_stream_t *stream,
+               bson_t *command,
+               int32_t max_wire_version)
 {
    bson_iter_t iter;
    bson_t change_stream_stage; /* { $changeStream: <change_stream_doc> } */
@@ -84,7 +103,6 @@ _make_command (mongoc_change_stream_t *stream, bson_t *command)
    bson_t pipeline;
    bson_t cursor_doc;
 
-   bson_init (command);
    if (stream->change_stream_type == MONGOC_CHANGE_STREAM_COLLECTION) {
       bson_append_utf8 (
          command, "aggregate", 9, stream->coll, (int) strlen (stream->coll));
@@ -99,16 +117,50 @@ _make_command (mongoc_change_stream_t *stream, bson_t *command)
    bson_append_document_begin (
       &change_stream_stage, "$changeStream", 13, &change_stream_doc);
    bson_concat (&change_stream_doc, stream->full_document);
-   if (!bson_empty (&stream->resume_token)) {
-      bson_concat (&change_stream_doc, &stream->resume_token);
-   }
-   /* Change streams spec: "startAtOperationTime and resumeAfter are mutually
-    * exclusive; if both startAtOperationTime and resumeAfter are set, the
-    * server will return an error. Drivers MUST NOT throw a custom error, and
-    * MUST defer to the server error." */
-   if (!_mongoc_timestamp_empty (&stream->operation_time)) {
-      _mongoc_timestamp_append (
-         &stream->operation_time, &change_stream_doc, "startAtOperationTime");
+
+   if (stream->resumed) {
+      /* Change stream spec: Resume Process */
+      if (!bson_empty (&stream->resume_token)) {
+         BSON_APPEND_DOCUMENT (
+            &change_stream_doc, "resumeAfter", &stream->resume_token);
+      } else if (!bson_empty (&stream->opts.startAfter)) {
+         BSON_APPEND_DOCUMENT (
+            &change_stream_doc, "resumeAfter", &stream->opts.startAfter);
+      } else if (!bson_empty (&stream->opts.resumeAfter)) {
+         BSON_APPEND_DOCUMENT (
+            &change_stream_doc, "resumeAfter", &stream->opts.resumeAfter);
+      } else if (!_mongoc_timestamp_empty (&stream->operation_time) &&
+                 max_wire_version >= 7) {
+         _mongoc_timestamp_append (&stream->operation_time,
+                                   &change_stream_doc,
+                                   "startAtOperationTime");
+      }
+   } else {
+      /* Change streams spec: "startAtOperationTime, resumeAfter, and startAfter
+       * are all mutually exclusive; if any two are set, the server will return
+       * an error. Drivers MUST NOT throw a custom error, and MUST defer to the
+       * server error." */
+      if (!bson_empty (&stream->opts.resumeAfter)) {
+         BSON_APPEND_DOCUMENT (
+            &change_stream_doc, "resumeAfter", &stream->opts.resumeAfter);
+
+         /* Update the cached resume token */
+         _set_resume_token (stream, &stream->opts.resumeAfter);
+      }
+
+      if (!bson_empty (&stream->opts.startAfter)) {
+         BSON_APPEND_DOCUMENT (
+            &change_stream_doc, "startAfter", &stream->opts.startAfter);
+
+         /* Update the cached resume token (take precedence over resumeAfter) */
+         _set_resume_token (stream, &stream->opts.startAfter);
+      }
+
+      if (!_mongoc_timestamp_empty (&stream->operation_time)) {
+         _mongoc_timestamp_append (&stream->operation_time,
+                                   &change_stream_doc,
+                                   "startAtOperationTime");
+      }
    }
 
    if (stream->change_stream_type == MONGOC_CHANGE_STREAM_CLIENT) {
@@ -175,7 +227,7 @@ _make_cursor (mongoc_change_stream_t *stream)
 
    BSON_ASSERT (stream);
    BSON_ASSERT (!stream->cursor);
-   _make_command (stream, &command);
+   bson_init (&command);
    bson_copy_to (&(stream->opts.extra), &command_opts);
    sd = mongoc_client_select_server (
       stream->client, false /* for_writes */, stream->read_prefs, &stream->err);
@@ -204,8 +256,9 @@ _make_cursor (mongoc_change_stream_t *stream)
       /* Create an implicit session. This session lsid must be the same for the
        * agg command and the subsequent getMores. Thus, this implicit session is
        * passed as if it were an explicit session to
-       * collection_read_command_with_opts and cursor_new_from_reply, but it is
-       * still implicit and its lifetime is owned by this change_stream_t. */
+       * mongoc_client_read_command_with_opts and
+       * _mongoc_cursor_change_stream_new, but it is still implicit and its
+       * lifetime is owned by this change_stream_t. */
       mongoc_session_opt_t *session_opts;
       session_opts = mongoc_session_opts_new ();
       mongoc_session_opts_set_causal_consistency (session_opts, false);
@@ -227,6 +280,8 @@ _make_cursor (mongoc_change_stream_t *stream)
       mongoc_read_concern_append (stream->read_concern, &command_opts);
    }
 
+   _make_command (stream, &command, max_wire_version);
+
    /* even though serverId has already been set, still pass the read prefs.
     * they are necessary for OP_MSG if sending to a secondary. */
    if (!mongoc_client_read_command_with_opts (stream->client,
@@ -264,20 +319,37 @@ _make_cursor (mongoc_change_stream_t *stream)
                          stream->batch_size);
    }
 
-   /* Change streams spec: "If neither startAtOperationTime nor resumeAfter are
-    * specified, and the max wire version is >= 7, and the initial aggregate
-    * command does not return a resumeToken (indicating no results), the
-    * ChangeStream MUST save the operationTime from the initial aggregate
-    * command when it returns." */
-   if (bson_empty (&stream->resume_token) &&
+   /* steals reply. */
+   stream->cursor =
+      _mongoc_cursor_change_stream_new (stream->client, &reply, &getmore_opts);
+
+   if (mongoc_cursor_error (stream->cursor, NULL)) {
+      goto cleanup;
+   }
+
+   /* Change stream spec: "When aggregate or getMore returns: If an empty batch
+    * was returned and a postBatchResumeToken was included, cache it." */
+   if (_mongoc_cursor_change_stream_end_of_batch (stream->cursor) &&
+       _mongoc_cursor_change_stream_has_post_batch_resume_token (
+          stream->cursor)) {
+      _set_resume_token (
+         stream,
+         _mongoc_cursor_change_stream_get_post_batch_resume_token (
+            stream->cursor));
+   }
+
+   /* Change stream spec: startAtOperationTime */
+   if (bson_empty (&stream->opts.resumeAfter) &&
+       bson_empty (&stream->opts.startAfter) &&
        _mongoc_timestamp_empty (&stream->operation_time) &&
-       max_wire_version >= 7 &&
-       bson_iter_init_find (&iter, &reply, "operationTime")) {
+       max_wire_version >= 7 && bson_empty (&stream->resume_token) &&
+       bson_iter_init_find (
+          &iter,
+          _mongoc_cursor_change_stream_get_reply (stream->cursor),
+          "operationTime") &&
+       BSON_ITER_HOLDS_TIMESTAMP (&iter)) {
       _mongoc_timestamp_set_from_bson (&stream->operation_time, &iter);
    }
-   /* steals reply. */
-   stream->cursor = mongoc_cursor_new_from_command_reply_with_opts (
-      stream->client, &reply, &getmore_opts);
 
 cleanup:
    bson_destroy (&command);
@@ -316,13 +388,8 @@ _change_stream_init (mongoc_change_stream_t *stream,
 
    stream->full_document = BCON_NEW ("fullDocument", stream->opts.fullDocument);
 
-   if (!bson_empty (&(stream->opts.resumeAfter))) {
-      bson_append_document (
-         &stream->resume_token, "resumeAfter", 11, &(stream->opts.resumeAfter));
-   }
-
    _mongoc_timestamp_set (&stream->operation_time,
-                          &(stream->opts.startAtOperationTime));
+                          &stream->opts.startAtOperationTime);
 
    stream->batch_size = stream->opts.batchSize;
    stream->max_await_time_ms = stream->opts.maxAwaitTimeMS;
@@ -415,10 +482,25 @@ _mongoc_change_stream_new_from_client (mongoc_client_t *client,
    return stream;
 }
 
+
+const bson_t *
+mongoc_change_stream_get_resume_token (mongoc_change_stream_t *stream)
+{
+   if (!bson_empty (&stream->resume_token)) {
+      return &stream->resume_token;
+   }
+
+   return NULL;
+}
+
+
 bool
 mongoc_change_stream_next (mongoc_change_stream_t *stream, const bson_t **bson)
 {
    bson_iter_t iter;
+   bson_t doc_resume_token;
+   uint32_t len;
+   const uint8_t *data;
    bool ret = false;
 
    BSON_ASSERT (stream);
@@ -440,20 +522,24 @@ mongoc_change_stream_next (mongoc_change_stream_t *stream, const bson_t **bson)
       }
 
       resumable = _is_resumable_error (err_doc);
-      if (resumable) {
+      while (resumable) {
          /* recreate the cursor. */
          mongoc_cursor_destroy (stream->cursor);
          stream->cursor = NULL;
+         stream->resumed = true;
          if (!_make_cursor (stream)) {
             goto end;
          }
-         if (!mongoc_cursor_next (stream->cursor, bson)) {
-            resumable =
-               !mongoc_cursor_error_document (stream->cursor, &err, &err_doc);
-            if (resumable) {
-               /* empty batch. */
-               goto end;
-            }
+         if (mongoc_cursor_next (stream->cursor, bson)) {
+            break;
+         }
+         if (!mongoc_cursor_error_document (stream->cursor, &err, &err_doc)) {
+            goto end;
+         }
+         if (err_doc) {
+            resumable = _is_resumable_error (err_doc);
+         } else {
+            resumable = false;
          }
       }
 
@@ -467,7 +553,8 @@ mongoc_change_stream_next (mongoc_change_stream_t *stream, const bson_t **bson)
 
    /* we have received documents, either from the first call to next or after a
     * resume. */
-   if (!bson_iter_init_find (&iter, *bson, "_id")) {
+   if (!bson_iter_init_find (&iter, *bson, "_id") ||
+       !BSON_ITER_HOLDS_DOCUMENT (&iter)) {
       bson_set_error (&stream->err,
                       MONGOC_ERROR_CURSOR,
                       MONGOC_ERROR_CHANGE_STREAM_NO_RESUME_TOKEN,
@@ -477,14 +564,27 @@ mongoc_change_stream_next (mongoc_change_stream_t *stream, const bson_t **bson)
    }
 
    /* copy the resume token. */
-   bson_reinit (&stream->resume_token);
-   BSON_APPEND_VALUE (
-      &stream->resume_token, "resumeAfter", bson_iter_value (&iter));
+   bson_iter_document (&iter, &len, &data);
+   BSON_ASSERT (bson_init_static (&doc_resume_token, data, len));
+   _set_resume_token (stream, &doc_resume_token);
+
    /* clear out the operation time, since we no longer need it to resume. */
    _mongoc_timestamp_clear (&stream->operation_time);
    ret = true;
 
 end:
+   /* Change stream spec: Updating the Cached Resume Token */
+   if (stream->cursor && !mongoc_cursor_error (stream->cursor, NULL) &&
+       _mongoc_cursor_change_stream_end_of_batch (stream->cursor) &&
+       _mongoc_cursor_change_stream_has_post_batch_resume_token (
+          stream->cursor)) {
+      _set_resume_token (
+         stream,
+         _mongoc_cursor_change_stream_get_post_batch_resume_token (
+            stream->cursor));
+   }
+
+
    /* Driver Sessions Spec: "When an implicit session is associated with a
     * cursor for use with getMore operations, the session MUST be returned to
     * the pool immediately following a getMore operation that indicates that the
