@@ -18,6 +18,7 @@
 #include <mongoc/mongoc.h>
 #include "mongoc/mongoc-collection-private.h"
 #include "mongoc/mongoc-util-private.h"
+#include "mongoc/mongoc-uri-private.h"
 
 #include "json-test.h"
 #include "json-test-operations.h"
@@ -205,8 +206,15 @@ process_sdam_test_ismaster_responses (bson_t *phase,
       bson_iter_bson (&ismaster_field_iter, &response);
 
       /* send ismaster through the topology description's handler */
+      capture_logs (true);
       mongoc_topology_description_handle_ismaster (
          td, sd->id, &response, 1, NULL);
+      if (td->servers->items_len == 0) {
+         ASSERT_CAPTURED_LOG ("topology",
+                              MONGOC_LOG_LEVEL_WARNING,
+                              "Last server removed from topology");
+      }
+      capture_logs (false);
    }
 }
 
@@ -620,9 +628,8 @@ get_bson_from_json_file (char *filename)
    return data;
 }
 
-
-static int
-check_scenario_version (const bson_t *scenario)
+static bool
+check_version_info (const bson_t *scenario, bool print_reason)
 {
    const char *s;
    char *padded;
@@ -635,9 +642,8 @@ check_scenario_version (const bson_t *scenario)
       test_version = test_framework_str_to_version (padded);
       bson_free (padded);
       server_version = test_framework_get_server_version ();
-
       if (server_version > test_version) {
-         if (test_suite_debug_output ()) {
+         if (print_reason && test_suite_debug_output ()) {
             printf ("      SKIP, maxServerVersion=\"%s\"\n", s);
             fflush (stdout);
          }
@@ -650,9 +656,8 @@ check_scenario_version (const bson_t *scenario)
       s = bson_lookup_utf8 (scenario, "minServerVersion");
       test_version = test_framework_str_to_version (s);
       server_version = test_framework_get_server_version ();
-
       if (server_version < test_version) {
-         if (test_suite_debug_output ()) {
+         if (print_reason && test_suite_debug_output ()) {
             printf ("      SKIP, minServerVersion=\"%s\"\n", s);
             fflush (stdout);
          }
@@ -661,10 +666,80 @@ check_scenario_version (const bson_t *scenario)
       }
    }
 
-   /* server version is ok, don't skip the test */
+   if (bson_has_field (scenario, "topology")) {
+      bson_iter_t iter;
+      bson_t topology;
+      char *current_topology;
+
+      BSON_ASSERT (bson_iter_init_find (&iter, scenario, "topology"));
+      BSON_ASSERT (BSON_ITER_HOLDS_ARRAY (&iter));
+
+      bson_iter_bson (&iter, &topology);
+
+      /* Determine cluster type */
+      if (test_framework_is_mongos ()) {
+         current_topology = "sharded";
+      } else if (test_framework_is_replset ()) {
+         current_topology = "replicaset";
+      } else {
+         current_topology = "single";
+      }
+
+      bson_iter_init (&iter, &topology);
+      while (bson_iter_next (&iter)) {
+         const char *test_topology;
+
+         BSON_ASSERT (BSON_ITER_HOLDS_UTF8 (&iter));
+         test_topology = bson_iter_utf8 (&iter, NULL);
+
+         if (strcmp (test_topology, current_topology) == 0) {
+            return true;
+         }
+      }
+
+      /* If we didn't match any of the listed topologies, skip */
+      if (print_reason && test_suite_debug_output ()) {
+         printf ("     SKIP, test topologies do not match current %s setup\n",
+                 current_topology);
+         fflush (stdout);
+      }
+
+      return false;
+   }
+
    return true;
 }
 
+static bool
+check_scenario_version (const bson_t *scenario)
+{
+   /* version info can be nested inside "runOn" array */
+   if (bson_has_field (scenario, "runOn")) {
+      bson_iter_t iter;
+      bson_t run_on;
+      bson_t version_info;
+
+      bson_lookup_doc (scenario, "runOn", &run_on);
+      BSON_ASSERT (bson_iter_init (&iter, &run_on));
+
+      while (bson_iter_next (&iter)) {
+         bson_iter_bson (&iter, &version_info);
+         if (check_version_info (&version_info, false)) {
+            return true;
+         }
+      }
+
+      if (test_suite_debug_output ()) {
+         printf ("      SKIP, no matching topologies in runOn\n");
+         fflush (stdout);
+      }
+
+      return false;
+   }
+
+   return check_version_info (scenario, true);
+}
+
 
 static int
 check_test_version (const bson_t *test)
@@ -780,50 +855,24 @@ check_topology_type (const bson_t *test)
 }
 
 
-/* insert the documents in a spec test scenario's "data" array */
 static void
-insert_data (mongoc_collection_t *collection, const bson_t *scenario)
+_insert_data (mongoc_collection_t *collection, bson_t *documents)
 {
-   mongoc_client_t *client;
-   mongoc_database_t *db;
-   mongoc_collection_t *tmp_collection;
-   bool r;
-   bson_error_t error;
-   mongoc_bulk_operation_t *bulk;
-   bson_t documents;
    bson_iter_t iter;
-   uint32_t server_id;
+   mongoc_bulk_operation_t *bulk;
    bson_t *majority = tmp_bson ("{'writeConcern': {'w': 'majority'}}");
+   bool r;
+   uint32_t server_id;
+   bson_error_t error;
 
-   /* clear data using a fresh client not configured with retryWrites etc. */
-   client = test_framework_client_new ();
-   db = mongoc_client_get_database (client, collection->db);
-   tmp_collection = mongoc_database_get_collection (db, collection->collection);
    mongoc_collection_delete_many (
-      tmp_collection, tmp_bson ("{}"), majority, NULL, NULL);
+      collection, tmp_bson ("{}"), majority, NULL, NULL);
 
-   mongoc_collection_destroy (tmp_collection);
-   /* ignore failure if it already exists */
-   tmp_collection = mongoc_database_create_collection (
-      db, collection->collection, majority, &error);
-
-   if (tmp_collection) {
-      mongoc_collection_destroy (tmp_collection);
-   }
-
-   mongoc_database_destroy (db);
-   mongoc_client_destroy (client);
-
-   if (!bson_has_field (scenario, "data")) {
+   if (!bson_count_keys (documents)) {
       return;
    }
 
-   bson_lookup_doc (scenario, "data", &documents);
-   if (!bson_count_keys (&documents)) {
-      return;
-   }
-
-   bson_iter_init (&iter, &documents);
+   bson_iter_init (&iter, documents);
    bulk =
       mongoc_collection_create_bulk_operation_with_opts (collection, majority);
 
@@ -846,11 +895,73 @@ insert_data (mongoc_collection_t *collection, const bson_t *scenario)
 }
 
 
+/* insert the documents in a spec test scenario's "data" array */
+static void
+insert_data (const char *db_name,
+             const char *collection_name,
+             const bson_t *scenario)
+{
+   mongoc_client_t *client;
+   mongoc_database_t *db;
+   mongoc_collection_t *collection;
+   mongoc_collection_t *tmp_collection;
+   bson_error_t error;
+   bson_t documents;
+   bson_iter_t iter;
+   bson_t *majority = tmp_bson ("{'writeConcern': {'w': 'majority'}}");
+
+   /* use a fresh client to prepare the collection */
+   client = test_framework_client_new ();
+
+   db = mongoc_client_get_database (client, db_name);
+   collection = mongoc_database_get_collection (db, collection_name);
+   mongoc_collection_delete_many (
+      collection, tmp_bson ("{}"), majority, NULL, NULL);
+
+   /* ignore failure if it already exists */
+   tmp_collection =
+      mongoc_database_create_collection (db, collection_name, majority, &error);
+
+   if (tmp_collection) {
+      mongoc_collection_destroy (tmp_collection);
+   }
+
+   if (!bson_has_field (scenario, "data")) {
+      goto DONE;
+   }
+
+   bson_iter_init_find (&iter, scenario, "data");
+
+   if (BSON_ITER_HOLDS_ARRAY (&iter)) {
+      bson_lookup_doc (scenario, "data", &documents);
+      _insert_data (collection, &documents);
+   } else {
+      /* go through collection: [] */
+      bson_iter_recurse (&iter, &iter);
+      while (bson_iter_next (&iter)) {
+         bson_t collection_documents;
+
+         mongoc_collection_destroy (collection);
+         collection =
+            mongoc_database_get_collection (db, bson_iter_key (&iter));
+         bson_iter_bson (&iter, &collection_documents);
+         _insert_data (collection, &collection_documents);
+      }
+   }
+
+DONE:
+   mongoc_database_destroy (db);
+   mongoc_collection_destroy (collection);
+   mongoc_client_destroy (client);
+}
+
+
 static void
 check_outcome_collection (mongoc_collection_t *collection, bson_t *test)
 {
    bson_t data;
    bson_iter_t iter;
+   mongoc_read_concern_t *rc;
    mongoc_cursor_t *cursor;
    bson_t query = BSON_INITIALIZER;
    mongoc_read_prefs_t *prefs = mongoc_read_prefs_new (MONGOC_READ_PRIMARY);
@@ -858,6 +969,15 @@ check_outcome_collection (mongoc_collection_t *collection, bson_t *test)
    bson_lookup_doc (test, "outcome.collection.data", &data);
    ASSERT (bson_iter_init (&iter, &data));
 
+   rc = mongoc_read_concern_new ();
+
+   /* If the collection has had its read_concern set by a test,
+      make sure it's set to LOCAL for this check. */
+   if (mongoc_read_concern_get_level (collection->read_concern)) {
+      mongoc_read_concern_set_level (rc, MONGOC_READ_CONCERN_LEVEL_LOCAL);
+      mongoc_collection_set_read_concern (collection, rc);
+   }
+
    cursor = mongoc_collection_find_with_opts (collection, &query, NULL, prefs);
 
    while (bson_iter_next (&iter)) {
@@ -874,6 +994,7 @@ check_outcome_collection (mongoc_collection_t *collection, bson_t *test)
    bson_destroy (&data);
    mongoc_read_prefs_destroy (prefs);
    bson_destroy (&query);
+   mongoc_read_concern_destroy (rc);
    mongoc_cursor_destroy (cursor);
 }
 
@@ -905,24 +1026,18 @@ execute_test (const json_test_config_t *config,
       collection->client->topology, MONGOC_SS_WRITE, NULL, &error);
    ASSERT_OR_PRINT (server_id, error);
 
-   if (bson_has_field (test, "failPoint")) {
-      bson_t command;
-      bool r;
-
-      bson_lookup_doc (test, "failPoint", &command);
-      ASSERT_CMPSTR (_mongoc_get_command_name (&command), "configureFailPoint");
-      r = mongoc_client_command_simple_with_server_id (
-         client, "admin", &command, NULL, server_id, NULL, &error);
-      ASSERT_OR_PRINT (r, error);
-   }
-
    json_test_ctx_init (&ctx, test, client, db, collection, config);
-   set_apm_callbacks (&ctx, collection->client);
 
    if (config->before_test_cb) {
       config->before_test_cb (&ctx, test);
    }
 
+   if (bson_has_field (test, "failPoint")) {
+      activate_fail_point (client, server_id, test, "failPoint");
+   }
+
+   set_apm_callbacks (&ctx, collection->client);
+
    json_test_operations (&ctx, test);
 
    if (config->after_test_cb) {
@@ -958,6 +1073,27 @@ execute_test (const json_test_config_t *config,
 }
 
 
+void
+activate_fail_point (mongoc_client_t *client,
+                     const uint32_t server_id,
+                     const bson_t *test,
+                     const char *key)
+{
+   bson_t command;
+   bson_error_t error;
+   bool r;
+
+   BSON_ASSERT (client);
+   BSON_ASSERT (server_id);
+
+   bson_lookup_doc (test, key, &command);
+
+   ASSERT_CMPSTR (_mongoc_get_command_name (&command), "configureFailPoint");
+   r = mongoc_client_command_simple_with_server_id (
+      client, "admin", &command, NULL, server_id, NULL, &error);
+   ASSERT_OR_PRINT (r, error);
+}
+
 /*
  *-----------------------------------------------------------------------
  *
@@ -1040,6 +1176,12 @@ set_uri_opts_from_bson (mongoc_uri_t *uri, const bson_t *opts)
       } else if (!strcmp (bson_iter_key (&iter), "retryWrites")) {
          mongoc_uri_set_option_as_bool (
             uri, "retryWrites", bson_iter_bool (&iter));
+      } else if (!strcmp (bson_iter_key (&iter), "heartbeatFrequencyMS")) {
+         mongoc_uri_set_option_as_int32 (
+            uri, "heartbeatFrequencyMS", bson_iter_int32 (&iter));
+      } else if (!strcmp (bson_iter_key (&iter), "retryReads")) {
+         mongoc_uri_set_option_as_bool (
+            uri, "retryReads", bson_iter_bool (&iter));
       } else {
          MONGOC_ERROR ("Unsupported clientOptions field \"%s\" in %s",
                        bson_iter_key (&iter),
@@ -1050,6 +1192,44 @@ set_uri_opts_from_bson (mongoc_uri_t *uri, const bson_t *opts)
 }
 
 
+static bool
+_should_skip_due_to_server_39704 (const bson_t *test)
+{
+   const char *desc = bson_lookup_utf8 (test, "description");
+   const char *bad_tests[] = {
+      "only first countDocuments includes readConcern",
+      "only first find includes readConcern",
+      "only first aggregate includes readConcern",
+      "only first distinct includes readConcern",
+      "only first runCommand includes readConcern",
+      "transaction options inherited from defaultTransactionOptions",
+      "startTransaction options override defaults",
+      "defaultTransactionOptions override client options",
+      "readConcern snapshot in startTransaction options",
+      "withTransaction inherits transaction options from "
+      "defaultTransactionOptions",
+      "withTransaction explicit transaction options",
+      "withTransaction explicit transaction options override "
+      "defaultTransactionOptions",
+      "withTransaction explicit transaction options override client options"};
+   int i;
+
+   /* Only an issue for sharded clusters. */
+   if (!test_framework_is_mongos ()) {
+      return false;
+   }
+
+   for (i = 0; i < sizeof (bad_tests) / sizeof (char *); i++) {
+      if (0 == strcmp (desc, bad_tests[i])) {
+         return true;
+      }
+   }
+
+
+   return false;
+}
+
+
 /*
  *-----------------------------------------------------------------------
  *
@@ -1101,6 +1281,7 @@ run_json_general_test (const json_test_config_t *config)
       uint32_t server_id;
       bson_error_t error;
       bool r;
+      bson_iter_t uri_iter;
 
       ASSERT (BSON_ITER_HOLDS_DOCUMENT (&tests_iter));
       bson_iter_bson (&tests_iter, &test);
@@ -1114,9 +1295,38 @@ run_json_general_test (const json_test_config_t *config)
          continue;
       }
 
+      if (bson_has_field (&test, "skipReason")) {
+         fprintf (stderr,
+                  " - %s SKIPPED, reason: %s\n",
+                  description,
+                  bson_lookup_utf8 (&test, "skipReason"));
+         continue;
+      }
+
+      if (_should_skip_due_to_server_39704 (&test)) {
+         fprintf (stderr,
+                  " - %s SKIPPED, reason: SERVER-39704 causes sharded tests to "
+                  "fail when using readConcern: snapshot\n",
+                  description);
+         continue;
+      }
+
       bson_free (selected_test);
 
       uri = test_framework_get_uri ();
+
+      /* If we are using multiple mongos, hardcode them in, for now, but keep
+       * the other URI components (CDRIVER-3285) */
+      if (bson_iter_init_find (&uri_iter, &test, "useMultipleMongoses") &&
+          bson_iter_as_bool (&uri_iter)) {
+         ASSERT_OR_PRINT (
+            mongoc_uri_upsert_host_and_port (uri, "localhost:27017", &error),
+            error);
+         ASSERT_OR_PRINT (
+            mongoc_uri_upsert_host_and_port (uri, "localhost:27018", &error),
+            error);
+      }
+
       if (bson_iter_init_find (&client_opts_iter, &test, "clientOptions")) {
          bson_t client_opts;
 
@@ -1151,9 +1361,10 @@ run_json_general_test (const json_test_config_t *config)
          MONGOC_WARNING ("Error in killAllSessions: %s", error.message);
       }
 
+      insert_data (db_name, collection_name, scenario);
+
       db = mongoc_client_get_database (client, db_name);
       collection = mongoc_database_get_collection (db, collection_name);
-      insert_data (collection, scenario);
       execute_test (config, client, db, collection, &test);
 
       mongoc_collection_destroy (collection);
