@@ -137,6 +137,8 @@ test_handshake_install (TestSuite *suite);
 extern void
 test_queue_install (TestSuite *suite);
 extern void
+test_primary_stepdown_install (TestSuite *suite);
+extern void
 test_read_concern_install (TestSuite *suite);
 extern void
 test_read_write_concern_install (TestSuite *suite);
@@ -145,6 +147,8 @@ test_read_prefs_install (TestSuite *suite);
 extern void
 test_retryable_writes_install (TestSuite *suite);
 extern void
+test_retryable_reads_install (TestSuite *suite);
+extern void
 test_rpc_install (TestSuite *suite);
 extern void
 test_samples_install (TestSuite *suite);
@@ -189,6 +193,8 @@ test_util_install (TestSuite *suite);
 extern void
 test_version_install (TestSuite *suite);
 extern void
+test_with_transaction_install (TestSuite *suite);
+extern void
 test_write_command_install (TestSuite *suite);
 extern void
 test_write_concern_install (TestSuite *suite);
@@ -1769,43 +1775,65 @@ test_framework_get_max_wire_version (int64_t *max_version)
    bson_destroy (&reply);
 }
 
+static bool
+_test_framework_has_auth (void)
+{
+   char *user;
+
+#ifndef MONGOC_ENABLE_SSL
+   /* requires SSL for SCRAM implementation, can't test auth */
+   return false;
+#endif
+
+   /* checks if the MONGOC_TEST_USER env var is set */
+   user = test_framework_get_admin_user ();
+   bson_free (user);
+   if (user) {
+      return true;
+   } else {
+      return false;
+   }
+}
+
 
 int
 test_framework_skip_if_auth (void)
 {
-   char *user;
-
    if (!TestSuite_CheckLive ()) {
       return 0;
    }
 
-   /* run tests if the MONGOC_TEST_USER env var is not set */
-   user = test_framework_get_admin_user ();
-   bson_free (user);
-   return user ? 0 : 1;
+   if (_test_framework_has_auth ()) {
+      return 0;
+   }
+
+   return 1;
 }
 
 
 int
 test_framework_skip_if_no_auth (void)
 {
-   char *user;
-
    if (!TestSuite_CheckLive ()) {
       return 0;
    }
 
-#ifndef MONGOC_ENABLE_SSL
-   /* requires SSL for SCRAM implementation, can't test auth */
-   return 0;
-#endif
+   if (!_test_framework_has_auth ()) {
+      return 0;
+   }
 
-   /* run auth tests if the MONGOC_TEST_USER env var is set */
-   user = test_framework_get_admin_user ();
-   bson_free (user);
-   return user ? 1 : 0;
+   return 1;
 }
 
+static bool
+_test_framework_has_crypto (void)
+{
+#ifdef MONGOC_ENABLE_CRYPTO
+   return true;
+#else
+   return false;
+#endif
+}
 
 int
 test_framework_skip_if_no_sessions (void)
@@ -1814,6 +1842,10 @@ test_framework_skip_if_no_sessions (void)
       return 0;
    }
 
+   if (!_test_framework_has_crypto ()) {
+      return 0;
+   }
+
    return -1 != test_framework_session_timeout_minutes ();
 }
 
@@ -1832,13 +1864,10 @@ test_framework_skip_if_no_cluster_time (void)
 int
 test_framework_skip_if_crypto (void)
 {
-#ifdef MONGOC_ENABLE_CRYPTO
-   return 0;
-#else
-   return 1;
-#endif
+   return _test_framework_has_crypto () ? 0 : 1;
 }
 
+
 int
 test_framework_skip_if_no_crypto (void)
 {
@@ -1919,6 +1948,13 @@ test_framework_skip_if_no_txns (void)
       return 1;
    }
 
+   if (test_framework_skip_if_no_crypto () &&
+       test_framework_skip_if_no_sessions () &&
+       test_framework_skip_if_not_mongos () &&
+       test_framework_skip_if_max_wire_version_less_than_8 ()) {
+      return 1;
+   }
+
    /* transactions not supported, skip the test */
    return 0;
 }
@@ -2221,6 +2257,34 @@ test_framework_skip_if_compressors (void)
    return !test_framework_skip_if_no_compressors ();
 }
 
+int
+test_framework_skip_if_no_failpoint (void)
+{
+   mongoc_client_t *client;
+   bool ret;
+   bson_error_t error;
+
+   client = test_framework_client_new ();
+   mongoc_client_set_error_api (client, MONGOC_ERROR_API_VERSION_2);
+   ret = mongoc_client_command_simple (
+      client,
+      "admin",
+      tmp_bson ("{'configureFailPoint': 'failCommand', 'mode': 'off', 'data': "
+                "{'errorCode': 10107, 'failCommands': ['count']}}"),
+      NULL,
+      NULL,
+      &error);
+   mongoc_client_destroy (client);
+
+   if (!ret) {
+      return 0; /* do not proceed */
+   }
+
+   /* proceed. */
+   return 1;
+}
+
+
 void
 test_framework_resolve_path (const char *path, char *resolved)
 {
@@ -2312,10 +2376,12 @@ main (int argc, char *argv[])
    test_log_install (&suite);
    test_matcher_install (&suite);
    test_queue_install (&suite);
+   test_primary_stepdown_install (&suite);
    test_read_concern_install (&suite);
    test_read_write_concern_install (&suite);
    test_read_prefs_install (&suite);
    test_retryable_writes_install (&suite);
+   test_retryable_reads_install (&suite);
    test_rpc_install (&suite);
    test_socket_install (&suite);
    test_opts_install (&suite);
@@ -2339,6 +2405,7 @@ main (int argc, char *argv[])
    test_usleep_install (&suite);
    test_util_install (&suite);
    test_version_install (&suite);
+   test_with_transaction_install (&suite);
    test_write_concern_install (&suite);
 #ifdef MONGOC_ENABLE_SSL
    test_stream_tls_install (&suite);
