@@ -10,18 +10,25 @@
 #include "json-test-operations.h"
 
 
-static void
+static bool
 retryable_writes_test_run_operation (json_test_ctx_t *ctx,
                                      const bson_t *test,
                                      const bson_t *operation)
 {
    bool *explicit_session = (bool *) ctx->config->ctx;
+   bson_t reply;
+   bool res;
 
-   json_test_operation (ctx,
-                        test,
-                        operation,
-                        ctx->collection,
-                        *explicit_session ? ctx->sessions[0] : NULL);
+   res = json_test_operation (ctx,
+                              test,
+                              operation,
+                              ctx->collection,
+                              *explicit_session ? ctx->sessions[0] : NULL,
+                              &reply);
+
+   bson_destroy (&reply);
+
+   return res;
 }
 
 
@@ -36,6 +43,7 @@ test_retryable_writes_cb (bson_t *scenario)
    config.ctx = &explicit_session;
    config.run_operation_cb = retryable_writes_test_run_operation;
    config.scenario = scenario;
+   config.command_started_events_only = true;
    explicit_session = true;
    run_json_general_test (&config);
    explicit_session = false;
@@ -425,21 +433,123 @@ test_remove_unacknowledged (void)
 static void
 test_retry_no_crypto (void *ctx)
 {
+   mongoc_uri_t *uri;
    mongoc_client_t *client;
+   mongoc_client_pool_t *pool;
 
    capture_logs (true);
+
+   /* Test that no warning is logged if retryWrites is disabled. Warning logic
+    * is implemented in mongoc_topology_new, but test all public APIs that use
+    * the common code path. */
+   client = mongoc_client_new ("mongodb://localhost/?retryWrites=false");
+   BSON_ASSERT (client);
+   ASSERT_NO_CAPTURED_LOGS ("mongoc_client_new and retryWrites=false");
+   mongoc_client_destroy (client);
+
+   uri = mongoc_uri_new ("mongodb://localhost/?retryWrites=false");
+   BSON_ASSERT (uri);
+
+   client = mongoc_client_new_from_uri (uri);
+   BSON_ASSERT (client);
+   ASSERT_NO_CAPTURED_LOGS ("mongoc_client_new_from_uri and retryWrites=false");
+   mongoc_client_destroy (client);
+
+   pool = mongoc_client_pool_new (uri);
+   BSON_ASSERT (pool);
+   ASSERT_NO_CAPTURED_LOGS ("mongoc_client_pool_new and retryWrites=false");
+   mongoc_client_pool_destroy (pool);
+
+   mongoc_uri_destroy (uri);
+
+   /* Test that a warning is logged if retryWrites is enabled. */
    client = mongoc_client_new ("mongodb://localhost/?retryWrites=true");
    BSON_ASSERT (client);
-
    ASSERT_CAPTURED_LOG (
-      "retryWrites=true",
+      "mongoc_client_new and retryWrites=true",
       MONGOC_LOG_LEVEL_WARNING,
       "retryWrites not supported without an SSL crypto library");
-
    mongoc_client_destroy (client);
+
+   clear_captured_logs ();
+
+   uri = mongoc_uri_new ("mongodb://localhost/?retryWrites=true");
+   BSON_ASSERT (uri);
+
+   client = mongoc_client_new_from_uri (uri);
+   BSON_ASSERT (client);
+   ASSERT_CAPTURED_LOG (
+      "mongoc_client_new_from_uri and retryWrites=true",
+      MONGOC_LOG_LEVEL_WARNING,
+      "retryWrites not supported without an SSL crypto library");
+   mongoc_client_destroy (client);
+
+   clear_captured_logs ();
+
+   pool = mongoc_client_pool_new (uri);
+   BSON_ASSERT (pool);
+   ASSERT_CAPTURED_LOG (
+      "mongoc_client_pool_new and retryWrites=true",
+      MONGOC_LOG_LEVEL_WARNING,
+      "retryWrites not supported without an SSL crypto library");
+   mongoc_client_pool_destroy (pool);
+
+   mongoc_uri_destroy (uri);
 }
 
+static void
+test_unsupported_storage_engine_error (void)
+{
+   mock_rs_t *rs;
+   mongoc_client_t *client;
+   mongoc_collection_t *coll;
+   bson_t reply;
+   bson_error_t error;
+   future_t *future;
+   request_t *request;
+   mongoc_client_session_t *session;
+   bson_t opts;
+   const char *expected_msg = "This MongoDB deployment does not support "
+                              "retryable writes. Please add retryWrites=false "
+                              "to your connection string.";
 
+   rs = mock_rs_with_autoismaster (WIRE_VERSION_RETRY_WRITES, true, 0, 0);
+   mock_rs_run (rs);
+   client = mongoc_client_new_from_uri (mock_rs_get_uri (rs));
+   session = mongoc_client_start_session (client, NULL, &error);
+   ASSERT_OR_PRINT (session, error);
+   mongoc_client_set_error_api (client, MONGOC_ERROR_API_VERSION_2);
+   coll = mongoc_client_get_collection (client, "test", "test");
+   bson_init (&opts);
+   ASSERT_OR_PRINT (mongoc_client_session_append (session, &opts, &error),
+                    error);
+   /* findandmodify is retryable through mongoc_client_write_command_with_opts.
+    */
+   future = future_client_write_command_with_opts (
+      client,
+      "test",
+      tmp_bson ("{'findandmodify': 'coll' }"),
+      &opts,
+      &reply,
+      &error);
+   request = mock_rs_receives_request (rs);
+   mock_server_replies_simple (
+      request,
+      "{'ok': 0, 'code': 20, 'errmsg': 'Transaction numbers are great'}");
+   request_destroy (request);
+
+   BSON_ASSERT (!future_get_bool (future));
+   ASSERT_ERROR_CONTAINS (error, MONGOC_ERROR_SERVER, 20, expected_msg);
+   ASSERT_MATCH (&reply, "{'code': 20, 'errmsg': '%s'}", expected_msg);
+
+   bson_destroy (&opts);
+   mongoc_client_session_destroy (session);
+   bson_destroy (&reply);
+   future_destroy (future);
+   mongoc_collection_destroy (coll);
+   mongoc_client_destroy (client);
+   mock_rs_destroy (rs);
+}
 /*
  *-----------------------------------------------------------------------
  *
@@ -501,4 +611,9 @@ test_retryable_writes_install (TestSuite *suite)
                       NULL,
                       NULL,
                       test_framework_skip_if_crypto);
+   TestSuite_AddMockServerTest (
+      suite,
+      "/retryable_writes/unsupported_storage_engine_error",
+      test_unsupported_storage_engine_error,
+      test_framework_skip_if_no_crypto);
 }
