@@ -1,5 +1,5 @@
-#include <mongoc.h>
-#include <mongoc-util-private.h>
+#include <mongoc/mongoc.h>
+#include <mongoc/mongoc-util-private.h>
 
 #include "TestSuite.h"
 #include "test-conveniences.h"
